unit uDataset;

interface

uses
  Forms, SysUtils, Classes, uDefTri, uGeral, uConexaoR, uMetodos, uProcessaTela,
  uFuncao, RCom,
  uDefEstatR, uFrmAguarde;

type
  TArrayId = array of String;

  TTRIDataset = class
  private
  public
    Nome : String;
    Calculou_ParItem : Boolean;
    Calculou_ParHab : Boolean;

    Habilidades : array of double; // habilidades estimadas
    Itens       : TDoubleMatrix; // parametros dos itens estimados

    IDItems     : TArrayId;
    IDIndividuos : TArrayID;

    Metodo : TMetodo;

    procedure SaveToDisk(pArquivo:String);
    procedure SaveParametrosItensToDisk(pArquivo:String);
    procedure SaveHabilidadesToDisk(pArquivo:String);
    procedure Estimar_Parametros_Itens(pMetodo:TMetodo);
    procedure Estimar_Habilidades(pMetodo:TMetodo);
    procedure Define_Estruturas_Globais_R;
    procedure OpenFile(pFile:String);

    procedure ExecutaFuncao(pFuncao:TFuncao; pNro:Integer);
  end;


  TParImportacao = record
    Tem_Individuos : Boolean;
    Tem_Habilidades : Boolean;
    LinHeader : Boolean; // Primeira linha tem header?
  end;


  TTriDataSets = class(TComposicao)
    function GetDataSet(pIdx:Integer; var pDataSet:TTRIDataSet):Boolean;
    function GetDataSetByName(pNome:String):TTriDataSet;
    procedure Importar(pNome:String; pArquivo:String; pParametros:TParImportacao);
  end;


var
  TRIDataSets : TTRIDataSets;

implementation


procedure TTRIDataSet.SaveToDisk(pArquivo:String);
var
  vComando : String;
begin
  vComando := format('write.matrix(%s, file="%s", sep = ",")',
                     [Nome, DefFilePathR(pArquivo)]);
  ConexaoR.RCon.ENR(vComando);
end;


procedure TTriDataSet.SaveParametrosItensToDisk(pArquivo:String);
var
  vComando : String;
begin
  Define_Estruturas_Globais_R;
  
  vComando := 'aux <- t(rbind(items.names, items.numoptions))';
  ConexaoR.RCon.ENR(vComando);

  if Calculou_ParItem then // concatena as estimativas:
  begin
    vComando := 'aux <- cbind(aux, ' + Nome+'.items.estimatives)';
    ConexaoR.RCon.ENR(vComando);
  end;

  vComando := format('write.matrix(%s, file="%s", sep = ",")',
                     ['aux', DefFilePathR(pArquivo)]);

  ConexaoR.RCon.ENR(vComando);
end;


procedure TTriDataSet.SaveHabilidadesToDisk(pArquivo:String);
var
  vComando : String;
begin
  Define_Estruturas_Globais_R;

  vComando := 'aux <- dataset.resp';
  ConexaoR.RCon.ENR(vComando);
  
  if Calculou_ParHab then
  begin
    vComando := 'aux <- cbind(aux, ' + Nome+'.abilities)';
    ConexaoR.RCon.ENR(vComando);
    {vComando := format('write.matrix(%s, file="%s", sep = ",")',
                       [Nome+'.abilities', DefFilePathR(pArquivo)]);}

  end;

  vComando := format('write.matrix(%s, file="%s", sep = ",")',
                     ['aux', DefFilePathR(pArquivo)]);

  ConexaoR.RCon.ENR(vComando);
end;


procedure TTriDataSet.Define_Estruturas_Globais_R;
begin
  // LAYOUT DO ARQUIVO:
  // # primeira linha = nome do gabarito,A,B,D,E,....
  // # segunda linha = nmero de alternativas da questo correspondente separado por ','
  // # terceira linha = S se o item deve ser considerado na anlise ou
  // #                  N se no deve ser considerado na anlise separado por ','
  // # USE SEMPRE LETRAS MAIUSCULAS para 'S' e 'N'
  // # quarta linha = nomes dos itens separados por vrgulas
  // # demais linhas = respostas dos indivduos: primeira coluna  a identific. do

  // variveis "globais"

  // Arquivo original:
  ConexaoR.RCon.ENR('dataset.original <- ' + nome+'.original');

  ConexaoR.RCon.Enr('dataset.working <- ' + nome + '.working');

  // Matriz contendo:
  // 1a linha: Cabealho dos dados
  // demais linhas: identificao do avaliado e notas dos avaliados
  ConexaoR.RCon.ENR('dataset <- dataset.original[4:dim(dataset.original)[1],]');

  // Matriz contendo:
  // 1a. linha: respostas corretas
  // 2a. linha: quantidade de opes existentes em0 cada item
  // 3a. linha: S ou N. S para informar que o item deve ser processado
  ConexaoR.RCon.ENR('dataset.info <- dataset.original[1:3,]');

  // Vetor com as respostas corretas
  // Exemplos: items.rightanswers[1] para obter a resposta correta do primeiro item
  //           items.rightanswers[10] para obter a resposta correta do dcimo item
  ConexaoR.RCon.ENR('items.rightanswers <- dataset.info[1,2:ncol(dataset.info)]');

  // Vetor com o nmero de alternativas para cada item:
  // Exemplo: items.numoptions[1] para obter a quantidade de alternativas para o item 1

        // A operao "dataset.info[2,1:(ncol(dataset.info)-1)]" retorna um "list"
        // ao fazer o unlist, os elementos do vetor so convertidos para texto
        // por isso  necessrio reconverter para nmero com as.numeric
  ConexaoR.RCon.ENR('items.numoptions <- as.numeric( unlist(dataset.info[2,1:(ncol(dataset.info)-1)]) )');

  // Vetor com flag informando se o item deve ser processado ou no:
  // Exemplo: items.deal[10] se for "S"  para processar
  ConexaoR.RCon.ENR('items.deal <- dataset.info[3,1:(ncol(dataset.info)-1)]');

  // Vetor com os nomes dos itens:
  // Exemplo: items.names[10] contm o nome do 10o. item:
  ConexaoR.RCon.ENR('items.names <- dataset.original[4,2:ncol(dataset.original)]');

  // Varivel com a quantidade de itens:
  // Exemplo: "items.count" poderia ter 30 (30 questes)
  ConexaoR.RCon.ENR('items.count <- ncol(dataset.info)-1');

  // Vetor com os identificadores dos indivduos:
  // Exemplo: dataset.resp[40] contm o identificador do quadragsimo respondente
  ConexaoR.RCon.ENR('dataset.resp <- dataset.original[5:nrow(dataset.original),1]');

  // Varivel com a quantidade de respondentes:
  // Exemplo: "dataset.count" poderia ter 300
  ConexaoR.RCon.ENR('dataset.count <- length(dataset.resp)');

  // Matriz com as respostas dadas pelos respondentes
  // Exemplo: dataset.answers[1,32] : resposta dada pelo 1o. respondente ao 32o. item
  ConexaoR.RCon.ENR('dataset.answers <- dataset.original[5:nrow(dataset.original),2:ncol(dataset.original)]');

  // Matriz com TRUE se o respondente acertou o item ou FALSE se errou:
  // ConexaoR.RCon.ENR('dataset.answers <- dataset.original[5:nrow(dataset.original),2:ncol(dataset.original)]');
end;


procedure TTriDataSet.OpenFile(pFile:String);
var
  vCmd : String;
  vQtd : Integer;
begin
  vCmd := 'ReadTest("' + nome + '", "' + DefFilePathR(pFile) + '")';
  ConexaoR.RCon.ENR( vCmd );

  vCmd := nome + '.count';
  vQtd := ConexaoR.RCon.Evaluate(vCmd);

  setLength(IDIndividuos, vQtd);

  vCmd := nome + '.resp';
  ConexaoR.RCon.GetStringVector(vCmd, vQtd, IDIndividuos);

  vCmd := nome + '.items.count';
  vQtd := ConexaoR.RCon.Evaluate(vCmd);
  SetLength(IDItems, vQtd);

  vCmd := nome + '.items.names';
  ConexaoR.RCon.GetStringVector(vCmd, vQtd, IDItems);
end;


procedure TTriDataSet.ExecutaFuncao(pFuncao:TFuncao; pNro:Integer);
var
  vComando : String;
begin
  if Metodo = Nil then
    raise Exception.Create('Mtodo no selecionado');

  Define_Estruturas_Globais_R;

  // Passa para o programa em R "items.estimatives" e "item.par"
  vComando := 'items.estimatives  <- ' + Nome+'.items.estimatives';
  ConexaoR.RCon.ENR(vComando);

  if pFuncao.Tipo = tf_Itens then
  begin
    vComando := 'item.par <- items.estimatives[' + IntToStr(pNro) + ',]';
    ConexaoR.RCon.ENR(vComando);
  end
  else
  begin
    vComando := 'abilities <- ' + Nome + '.abilities';
    ConexaoR.RCon.ENR(vComando);

    // Passa para o programa em "R"
    vComando := 'ability <- ' + Nome+'.abilities[' + IntToStr(pNro) + ']';
    ConexaoR.RCon.ENR(vComando);
  end;

  ConexaoR.RCon.Enr('save.image("' + PathExecutavelR + 'before.Rdata")');

  ConexaoR.RCon.Enr('source("' + DefFilePathR(Metodo.DirTemp+pFuncao.Programa) + '")');

  ConexaoR.RCon.Enr('save.image("' + PathExecutavelR + 'after.Rdata")');
end;


procedure TTriDataSet.Estimar_Parametros_Itens(pMetodo:TMetodo);
var
  vAguarde : TFrmAguarde;
  vIdx : Integer;
  vElemento : TElemento;
  vComando : String;
  vNumCol,
  vNumLin : Integer;
  I : Integer;
begin
  Metodo := pMetodo;

  vAguarde := TFrmAguarde.Create(Nil);
  try
    vAguarde.Show;

    vIdx := 0;
    while pMetodo.Parametros.Modulo.GetElemento(vIdx,vElemento) do
    begin
      vComando := vElemento.Nome + ' <- ' + vElemento.Valor;
      try
        ConexaoR.RCon.ENR(vComando);
      except
        raise;
      end;
      Inc(vIdx);
    end;

    Define_Estruturas_Globais_R;

    ConexaoR.RCon.Enr('save.image("' + PathExecutavelR + 'before.Rdata")');

    ConexaoR.RCon.Enr('source("' + DefFilePathR(pMetodo.DirTemp+pMetodo.Parametros.Programa) + '")');

    ConexaoR.RCon.Enr('save.image("' + PathExecutavelR + 'after.Rdata")');

    // Esperado que os resultados sejam colocados em items.estimatives
    ConexaoR.RCon.ENR( Nome+'.items.estimatives <- items.estimatives');

    vNumLin := ConexaoR.RCon.Evaluate('nrow(items.estimatives)');
    vNumCol := ConexaoR.RCon.Evaluate('ncol(items.estimatives)');
    setLength(Itens, vNumLin);
    for I := 1 to vNumLin do
      SetLength(Itens[I-1], vNumCol);
    ConexaoR.RCon.GetDoubleMatrix('items.estimatives', vNumLin, vNumCol, Itens);

    Calculou_ParItem := True;
  finally
    vAguarde.Free;
  end;
end;


procedure TTriDataSet.Estimar_Habilidades(pMetodo:TMetodo);
var
  vAguarde : TFrmAguarde;
  vIdx : Integer;
  vElemento : TElemento;
  vComando : String;
  vQtdIndividuos : Integer;
begin
  Metodo := pMetodo;

  vAguarde := TFrmAguarde.Create(Nil);
  try
    vAguarde.Show;
    Application.ProcessMessages;

    vIdx := 0;
    while pMetodo.Habilidades.Modulo.GetElemento(vIdx,vElemento) do
    begin
      vComando := vElemento.Nome + ' <- ' + vElemento.Valor;
      try
        ConexaoR.RCon.ENR(vComando);
      except
        raise;
      end;
      Inc(vIdx);
    end;

    Define_Estruturas_Globais_R;

    ConexaoR.RCon.Enr('save.image("' + PathExecutavelR + 'before.Rdata")');

    ConexaoR.RCon.Enr('source("' + DefFilePathR(pMetodo.DirTemp+pMetodo.Habilidades.Programa) + '")');

    ConexaoR.RCon.Enr('save.image("' + PathExecutavelR + 'after.Rdata")');

    // Esperado que os resultados sejam colocados em Nome.abilities
    ConexaoR.RCon.ENR( Nome+'.abilities <- abilities');

    vQtdIndividuos := High(IDIndividuos)+1;
    SetLength(Habilidades, vQtdIndividuos);
    ConexaoR.RCon.GetDoubleVector('abilities', vQtdIndividuos, Habilidades);

    {
    vNumLin := ConexaoR.RCon.Evaluate('nrow(items.estimatives)');
    vNumCol := ConexaoR.RCon.Evaluate('ncol(items.estimatives)');
    setLength(Itens, vNumLin);
    for I := 1 to vNumLin do
      SetLength(Itens[I-1], vNumCol);
    ConexaoR.RCon.GetDoubleMatrix('items.estimatives', vNumLin, vNumCol, Itens);
    }

    Calculou_ParHab := True;
  finally
    vAguarde.Free;
  end;
end;


//******************************************************************************
function TTRiDataSets.GetDataSet(pIdx:Integer; var pDataSet:TTRIDataSet):Boolean;
begin
  GetItem(pidx, Result, TObject(pDataSet));
end;


function TTriDataSets.GetDataSetByName(pNome:String):TTriDataSet;
var
  vIdx : Integer;
  vDataSet : TTRIDataSet;
begin
  Result := Nil;
  vIdx := 0;
  while (vIdx < Count) and (Result = Nil) do
  begin
    if GetDataSet(vIdx, vDataSet)
    and(vDataSet.Nome=pNome) then
        Result := vDataSet
    else
      Inc(vIdx);
  end;
end;


procedure TTRIDataSets.Importar(pNome:String; pArquivo:String; pParametros:TParImportacao);
var
  vHeader    : String;
  vComando   : String;

  vDataSet   : TTRIDataSet;
begin
  if pParametros.Tem_Individuos then
  begin
    (*
    pArquivo := DefPathR(ExtractFilePath(pArquivo)) + ExtractFileName(pArquivo);

    if pParametros.LinHeader then
      vHeader := 'header=TRUE'
    else
      vHeader := 'header=FALSE';

    // "as.is = true" faz com que seja possvel ler cada clula com um par de argumentos da matrix ([x,y])
    // fill ?
    vComando := '%s <- read.table("%s", %s, fill=TRUE, sep=",", as.is=TRUE)';
    vComando := format(vComando,[ pNome, pArquivo, vHeader]);

    ConexaoR.RCon.Enr(vComando);
    *)

    vDataSet := TTRiDataSet.Create;
    vDataSet.Nome := pNome;
    Add(vDataSet);

    vDataset.OpenFile( pArquivo )
  end;
end;


initialization
TRIDataSets:= TTRIDataSets.Create;


finalization
TRIDataSets.Free;


end.
